/*
 * arch/mips/kernel/entry.S
 *
 * Copyright (C) 1994, 1995 Waldorf Electronics
 * written by Ralf Baechle and Andreas Busse
 */

/*
 * entry.S contains the system-call and fault low-level handling routines.
 * This also contains the timer-interrupt handler, as well as all interrupts
 * and faults that can result in a task-switch.
 */

#include <linux/sys.h>

#include <asm/asm.h>
#include <asm/segment.h>
#include <asm/mipsregs.h>
#include <asm/mipsconfig.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/stackframe.h>
#include <asm/pica.h>
#include <asm/processor.h>

/*
 * These are offsets into the task-struct.
 */
state		=  0
counter		=  4
priority	=  8
signal		= 12
blocked		= 16
flags		= 20
errno		= 24
exec_domain	= 60

ENOSYS		= 38

		.globl	ret_from_sys_call

		.text
		.set	noreorder
		.align 4
handle_bottom_half:
		/*
		 * If your assembler breaks on the next line it's
		 * time to update!
		 */
		lui	s0,%hi(C_LABEL(intr_count))
		lw	s1,%lo(C_LABEL(intr_count))(s0)
		mfc0	s3,CP0_STATUS		# Enable IRQs
		addiu	s2,s1,1
		sw	s2,%lo(C_LABEL(intr_count))(s0)
		ori	t0,s3,0x1f
		xori	t0,0x1e
		jal	C_LABEL(do_bottom_half)
		mtc0	t0,CP0_STATUS		# delay slot
		mtc0	s3,CP0_STATUS		# Restore old IRQ state
		j	9f
		sw	s1,%lo(C_LABEL(intr_count))(s0)	# delay slot

		.set	reorder
reschedule:
		la	ra,ret_from_sys_call
		j	C_LABEL(schedule)
		nop

		.align	5
		NESTED(handle_sys, FR_SIZE, sp)
		.set	noreorder
		.set	noat
		SAVE_ALL
		.set	at
		STI
		/*
		 * Compute return address. For now we assume that syscalls never
		 * appear in branch delay slots. For the Linux/MIPS standard
		 * libraries this assumption is always try.
		 */
		lw	t3,FR_EPC(sp)
		lw	s1,FR_REG2(sp)
		li	t0,-ENOSYS
		addiu	t3,4
		sw	t3,FR_EPC(sp)
		li	t2,NR_syscalls
		bge	s1,t2,ret_from_sys_call
		sw	t0,FR_REG2(sp)		# delay slot
		sll	s1,2
		lw	s1,C_LABEL(sys_call_table)(s1)
		lw	s0,C_LABEL(current)

		beqz	s1,ret_from_sys_call
		lw	t0,flags(s0)
		sll	t0,2			# PF_TRACESYS
		bltz	t0,1f
		sw	zero,errno(s0)		# delay slot

		lw	a0,FR_REG4(sp)
		lw	a1,FR_REG5(sp)
		lw	a2,FR_REG6(sp)
		lw	a3,FR_REG7(sp)
		lw	t0,FR_REG3(sp)
		jalr	s1			# do the real work
		sw	t0,PTRSIZE*4(sp)	# delay slot

		lw	t0,errno(s0)
		sw	v0,FR_REG2(sp)		# save the return value
		subu	t0,zero,t0		# t0 = -t0
		beqz	t0,ret_from_sys_call
		nop
		/*
		 * Fixme: should set error flag
		 */
		j	ret_from_sys_call
		sw	t0,FR_REG2(sp)		# delay slot

		.align	4
1:		jal	C_LABEL(syscall_trace)
		nop				# delay slot

		lw	a0,FR_REG4(sp)
		lw	a1,FR_REG5(sp)
		lw	a2,FR_REG6(sp)
		lw	a3,FR_REG7(sp)
		lw	t0,FR_REG3(sp)
		jalr	s1			# do the real work
		sw	t0,PTRSIZE*4(sp)	# delay slot

		lw	t0,errno(s0)
		sw	v0,FR_REG2(sp)		# save the return value
		subu	t0,zero,t0
		beqz	t0,1f
		nop				# delay slot
		sw	t1,FR_REG2(sp)
		/*
		 * Fixme: should set error flag
		 */
1:		jal	C_LABEL(syscall_trace)
		nop

		.align	4
ret_from_sys_call:
		lw	t0,C_LABEL(intr_count)		# bottom half
		bnez	t0,2f
9:
		lw	t0,C_LABEL(bh_mask)	# delay slot
		lw	t1,C_LABEL(bh_active)	# unused delay slot
		and	t0,t1
		bnez	t0,handle_bottom_half

		lw	t0,FR_STATUS(sp)	# returning to supervisor ?
		andi	t1,t0,0x10
		beqz	t1,2f

		mfc0	t0,CP0_STATUS		# delay slot
		lw	t1,C_LABEL(need_resched)
		ori	t0,0x1f			# enable irqs
		xori	t0,0x1e
		bnez	t1,reschedule
		mtc0	t0,CP0_STATUS		# delay slot

		lw	s0,C_LABEL(current)
		lw	t0,C_LABEL(task)
		lw	t1,state(s0)		# state
		beq	s0,t0,2f		# task[0] cannot have signals
		lw	t0,counter(s0)		# counter
		bnez	t1,reschedule		# state == 0 ?
		lw	a0,blocked(s0)
						# save blocked in a0 for
						# signal handling
		beqz	t0,reschedule		# counter == 0 ?
		lw	t0,signal(s0)
		nor	t1,zero,a0
		and	t1,t0,t1
		beqz	t1,skip_signal_return
		nop

		jal	C_LABEL(do_signal)
		move	a1,sp			# delay slot
		
skip_signal_return:
		.set	noreorder
		.set	noat
2:
		.globl	return
return:		RESTORE_ALL
		eret
		.set	at
		END(handle_sys)

/*
 * Beware: interrupt, fast_interrupt, bad_interrupt and spurious_interrupt
 * have unusal calling conventions!
 *
 * t1 - interrupt number
 * s2 - destroyed
 * return values:
 * v0 - return routine
 */
		.text
		.set	noreorder
		.set	at
		.align	5
		NESTED(interrupt, FR_SIZE, sp)
		move	s2,ra
		mfc0	t0,CP0_STATUS
		ori	t0,0x1f
		xori	t0,0x1e
		mtc0	t0,CP0_STATUS
		move	a0,t1
		jal	C_LABEL(do_IRQ)
		move	a1,sp				# delay slot
		mfc0	t0,CP0_STATUS
		ori	t0,1
		xori	t0,1
		la	v0,ret_from_sys_call
		jr	s2
		mtc0	t0,CP0_STATUS			# delay slot
		END(interrupt)

		.align	5
		NESTED(fast_interrupt, FR_SIZE, sp)
		move	s2,ra
		move	a0,t1
		jal	C_LABEL(do_fast_IRQ)
		move	a1,sp				# delay slot
		la	v0,return
		jr	s2
		nop					# delay slot
		END(fast_interrupt)

		LEAF(bad_interrupt)
		/*
		 * Don't return & unblock the pic
		 */
		j	return
		nop
		END(bad_interrupt)

		.align	5
		LEAF(spurious_interrupt)
		/*
		 * Nothing happened... (whistle)
		 */
		lui	t1,%hi(C_LABEL(spurious_count))
		lw	t0,%lo(C_LABEL(spurious_count))(t1)
		la	v0,return
		addiu	t0,1
		jr	ra
		sw	t0,%lo(C_LABEL(spurious_count))(t1)
		END(spurious_interrupt)

		.align	5
		NESTED(handle_tlbl, FR_SIZE, sp)
		.set	noreorder
		.set	noat
		/*
		 * Check whether this is a refill or an invalid exception
		 *
		 * NOTE: Some MIPS manuals say that the R4x00 sets the
		 * BadVAddr only when EXL == 0. This is wrong - BadVaddr
		 * is being set for all Reload, Invalid and Modified
		 * exceptions.
		 */
#if 0
	lui	k0,%hi(_buggy)
	lw	k0,%lo(_buggy)(k0)
	beqz	k0,1f
	nop
	SAVE_ALL
	PRINT("handle_tlbl()\n")
	mfc0	a1,CP0_BADVADDR
	PRINT("c0_badvaddr == %08x\n")
	jal	_dump_tlb
	nop
2:	j	2b
	nop
	RESTORE_ALL
1:	nop
#endif
		mfc0	k0,CP0_BADVADDR
		mfc0	k1,CP0_ENTRYHI
		ori	k0,0x1fff
		xori	k0,0x1fff
		andi	k1,0xff
		or	k0,k1
		mfc0	k1,CP0_ENTRYHI
		mtc0	k0,CP0_ENTRYHI
		nop					# for R4[04]00 pipeline
		nop
		nop
		tlbp
		nop					# for R4[04]00 pipeline
		nop
		mfc0	k0,CP0_INDEX
		srl	k0,31
		beqz	k0,invalid_tlbl
		mtc0	k1,CP0_ENTRYHI			# delay slot
		/*
		 * Not in tlb -> nested refill exception
		 * Load the missing entry and return. This is the most
		 * efficient way to regain the faulting address.
		 */

		/*
		 * Damn... The next nop is required on my R4400PC, but
		 * I don't know why - at least there is no documented
		 * reason as for the others :-(
		 */
		nop
		dmfc0	k1,CP0_CONTEXT
		dsra	k1,1
		lwu	k0,(k1)			# Never causes another exception
		lwu	k1,4(k1)
		dsrl	k0,6			# Convert to EntryLo format
		dsrl	k1,6			# Convert to EntryLo format
		dmtc0	k0,CP0_ENTRYLO0
		dmtc0	k1,CP0_ENTRYLO1
		nop				# for R4[04]00 pipeline
		tlbwr
		nop				# for R4[04]00 pipeline
		nop
		nop
		eret

		/*
		 * Handle invalid exception
		 *
		 * There are two possible causes for an invalid (tlbl)
		 * exception:
		 * 1) pages that have the present bit set but the valid bit
		 *    unset.
		 * 2) pages that don't exist
		 * Case one needs fast handling, therefore don't save
		 * registers yet.
		 *
		 * k0 now contains the bad virtual address.
		 */
invalid_tlbl:
		/*
		 * Remove entry so we don't need to care later
		 */
		mfc0	k0,CP0_INDEX
		lui	k1,0x0008
		or	k0,k1
		dsll	k0,13
		dmtc0	k0,CP0_ENTRYHI
		dmtc0	zero,CP0_ENTRYLO0
		dmtc0	zero,CP0_ENTRYLO1
		/*
		 * Test whether present bit in entry is set
		 */
		dmfc0	k0,CP0_BADVADDR
		tlbwi				# delayed, for R4[04]00 pipeline
		srl	k0,10
		lui	k1,%HI(TLBMAP)
		addu	k0,k1
		ori	k0,3
		xori	k0,3
		lw	k1,(k0)
		andi	k1,_PAGE_PRESENT
		beqz	k1,nopage_tlbl
		/*
		 * Present bit is set -> set valid and accessed bits
		 */
		lw	k1,(k0)				# delay slot
		ori	k1,_PAGE_ACCESSED
		sw	k1,(k0)
		eret

		/*
		 * Page doesn't exist. Lots of work which is less important
		 * for speed needs to be done, so hand it all over to the
		 * kernel memory management routines.
		 */
nopage_tlbl:
		SAVE_ALL
		.set	at
		STI
		/*
		 * Create a Intel-style errorcode
		 * Bit 0: P Present
		 *        0 == Page not in memory
		 *        1 == privilege violation
		 * Bit 1: R/W Read/Write
		 *        0 == ReadAccess
		 *        1 == WriteAccess
		 * Bit 2: U/S User/Supervisor
		 *        0 == User mode
		 *        1 == Kernel mode
		 *
		 * a0 (struct pt_regs *) regs
		 * a1 (unsigned long)    error_code
		 */
		lw	a1,FR_STATUS(sp)
		move	a0,sp
		srl	a1,4
		andi	a1,1
		jal	C_LABEL(do_page_fault)
		xori	a1,1				# delay slot
		j	ret_from_sys_call
		nop					# delay slot
		END(handle_tlbl)

		.text
		.align	5
		NESTED(handle_tlbs, FR_SIZE, sp)
#if 1
	.set	noat
	lui	k0,%hi(_buggy)
	lw	k0,%lo(_buggy)(k0)
	beqz	k0,1f
	nop
	SAVE_ALL
	.set	at
	PRINT("handle_tlbs()\n")
	mfc0	a1,CP0_BADVADDR
	PRINT("c0_badvaddr == %08x\n")
	jal	_dump_tlb
	nop
2:	j	2b
	nop
	.set	noat
	RESTORE_ALL
1:	nop
	.set	at
#endif
		.set	noreorder
		.set	noat
		/*
		 * It is impossible that is a nested reload exception.
		 * Therefore this must be a invalid exception.
		 * Two possible cases:
		 * 1) Page not used yet
		 * 2) Page doesn't exist yet. Let the kernel handle the trouble.
		 *
		 * Test whether present bit in entry is set
		 */
		dmfc0	k0,CP0_BADVADDR
		srl	k0,10
		lui	k1,%HI(TLBMAP)
		addu	k0,k1
		ori	k0,3
		xori	k0,3
		lw	k1,(k0)
		andi	k1,(_PAGE_PRESENT|_PAGE_RW)
		beqz	k1,nopage_tlbs
		/*
		 * Present and writable bits set -> set accessed and dirty bits.
		 */
		lw	k1,(k0)				# delay slot
		ori	k1,k1,(_PAGE_ACCESSED|_PAGE_DIRTY)
		sw	k1,(k0)
		/*
		 * Now reload the entry into the tlb
		 */
		ori	k0,0x1000
		xori	k0,0x1000
		lw	k1,4(k0)
		lw	k0,(k0)
		srl	k0,6
		srl	k1,6
		dmtc0	k0,CP0_ENTRYLO0
		dmtc0	k1,CP0_ENTRYLO1
		nop				# for R4[04]00 pipeline
		tlbwi
		eret

		/*
		 * Page doesn't exist. Lots of work which is less important
		 * for speed needs to be done, so hand it all over to the
		 * kernel memory management routines.
		 */
nopage_tlbs:
nowrite_mod:
#if 1
	lui	k0,%hi(_buggy)
	lw	k0,%lo(_buggy)(k0)
	beqz	k0,1f
	nop
	SAVE_ALL
	PRINT("handle_mod()\n")
	mfc0	a1,CP0_BADVADDR
	PRINT("c0_badvaddr == %08x\n")
	jal	_dump_tlb
	nop
2:	j	2b
	nop
	RESTORE_ALL
1:	nop
#endif
		/*
		 * Remove entry so we don't need to care later
		 */
		mfc0	k0,CP0_INDEX
		lui	k1,0x0008
		or	k0,k1
		dsll	k0,13
		dmtc0	k0,CP0_ENTRYHI
		dmtc0	zero,CP0_ENTRYLO0
		dmtc0	zero,CP0_ENTRYLO1
		nop				# for R4[04]00 pipeline
		tlbwi
		SAVE_ALL
		.set	at
		STI
		/*
		 * Create a Intel-style errorcode
		 * Bit 0: P Present
		 *        0 == Page not in memory
		 *        1 == privilege violation
		 * Bit 1: R/W Read/Write
		 *        0 == ReadAccess
		 *        1 == WriteAccess
		 * Bit 2: U/S User/Supervisor
		 *        0 == User mode
		 *        1 == Kernel mode
		 *
		 * a0 (struct pt_regs *) regs
		 * a1 (unsigned long)    error_code
		 */
		lw	a1,FR_STATUS(sp)
		move	a0,sp
		srl	a1,4
		andi	a1,1
		jal	C_LABEL(do_page_fault)
		xori	a1,3				# branch delay slot
		j	ret_from_sys_call
		nop					# branch delay slot
		END(handle_tlbs)

		.align	5
		NESTED(handle_mod, FR_SIZE, sp)
		.set	noreorder
		.set	noat
		/*
		 * Two possible cases:
		 * 1) Page is rw but not dirty -> set dirty and return
		 * 2) Page is not rw -> call C handler
		 */
		dmfc0	k0,CP0_BADVADDR
		srl	k0,10
		lui	k1,%HI(TLBMAP)
		addu	k0,k1
		ori	k0,3
		xori	k0,3
		lw	k1,(k0)
		andi	k1,_PAGE_RW
		beqz	k1,nopage_tlbs
		/*
		 * Present and writable bits set -> set accessed and dirty bits.
		 */
		lw	k1,(k0)				# delay slot
		ori	k1,(_PAGE_ACCESSED|_PAGE_DIRTY)
		sw	k1,(k0)
		/*
		 * Now reload the entry into the tlb
		 */
		ori	k0,0x1000
		xori	k0,0x1000
		lw	k1,4(k0)
		lw	k0,(k0)
		srl	k0,6
		srl	k1,6
		dmtc0	k0,CP0_ENTRYLO0
		dmtc0	k1,CP0_ENTRYLO1
		nop				# for R4[04]00 pipeline
		tlbwi
		nop				# for R4[04]00 pipeline
		nop
		nop
		eret
		END(handle_mod)

/*
 * Build a default exception handler for the other R4x00 exceptions
 */
#define BUILD_HANDLER(exception)                                        \
		.align	5;                                              \
		NESTED(handle_##exception, FR_SIZE, sp);                \
		.set	noreorder;                                      \
		.set	noat;                                           \
		SAVE_ALL;                                               \
		.set	at;                                             \
		li	t0,LED_2;                                       \
		sw	t0,PICA_LED;                                    \
		la	a1,8f;                                          \
		TEXT	(#exception);                                   \
		mfc0	a2,CP0_EPC;                                     \
		PRINT("Got %s at %08x.\n");                             \
		lw	a1,FR_EPC(sp);                                  \
		PRINT("Last do_IRQ EPC was %08x\n");                    \
		move	a0,sp;                                          \
		jal	C_LABEL(show_regs);                             \
		nop;                                                    \
1:		j	1b;                                             \
		nop;                                                    \
		STI;                                                    \
		li	t0,-1;                                          \
		sw	t0,FR_ORIG_REG2(sp);                            \
		jal	C_LABEL(do_##exception);                        \
		move	a0,sp;                   /* delay slot */       \
		li	t0,LED_2;                                       \
		sw	t0,PICA_LED;                                    \
		j	ret_from_sys_call;                              \
		nop;                             /* delay slot */       \
		END(handle_##exception)

		BUILD_HANDLER(adel)
		BUILD_HANDLER(ades)
		BUILD_HANDLER(ibe)
		BUILD_HANDLER(dbe)
		BUILD_HANDLER(ov)
		BUILD_HANDLER(fpe)
		BUILD_HANDLER(bp)
		BUILD_HANDLER(tr)
		BUILD_HANDLER(ri)
		BUILD_HANDLER(cpu)
		BUILD_HANDLER(vcei)
		BUILD_HANDLER(vced)
		BUILD_HANDLER(watch)
		BUILD_HANDLER(reserved)

/*
 * Exception handler table with 32 entries.
 * This might be extended to handle software exceptions
 */
		.bss
		.align	2
		EXPORT(exception_handlers)
		.fill	32,4,0

/*
 * Table of syscalls
 */
		.data
		EXPORT(sys_call_table)
		PTR	C_LABEL(sys_setup)		/* 0 */
		PTR	C_LABEL(sys_exit)
		PTR	C_LABEL(sys_fork)
		PTR	C_LABEL(sys_read)
		PTR	C_LABEL(sys_write)
		PTR	C_LABEL(sys_open)		/* 5 */
		PTR	C_LABEL(sys_close)
		PTR	C_LABEL(sys_waitpid)
		PTR	C_LABEL(sys_creat)
		PTR	C_LABEL(sys_link)
		PTR	C_LABEL(sys_unlink)		/* 10 */
		PTR	C_LABEL(sys_execve)
		PTR	C_LABEL(sys_chdir)
		PTR	C_LABEL(sys_time)
		PTR	C_LABEL(sys_mknod)
		PTR	C_LABEL(sys_chmod)		/* 15 */
		PTR	C_LABEL(sys_chown)
		PTR	C_LABEL(sys_break)
		PTR	C_LABEL(sys_stat)
		PTR	C_LABEL(sys_lseek)
		PTR	C_LABEL(sys_getpid)		/* 20 */
		PTR	C_LABEL(sys_mount)
		PTR	C_LABEL(sys_umount)
		PTR	C_LABEL(sys_setuid)
		PTR	C_LABEL(sys_getuid)
		PTR	C_LABEL(sys_stime)		/* 25 */
		PTR	C_LABEL(sys_ptrace)
		PTR	C_LABEL(sys_alarm)
		PTR	C_LABEL(sys_fstat)
		PTR	C_LABEL(sys_pause)
		PTR	C_LABEL(sys_utime)		/* 30 */
		PTR	C_LABEL(sys_stty)
		PTR	C_LABEL(sys_gtty)
		PTR	C_LABEL(sys_access)
		PTR	C_LABEL(sys_nice)
		PTR	C_LABEL(sys_ftime)		/* 35 */
		PTR	C_LABEL(sys_sync)
		PTR	C_LABEL(sys_kill)
		PTR	C_LABEL(sys_rename)
		PTR	C_LABEL(sys_mkdir)
		PTR	C_LABEL(sys_rmdir)		/* 40 */
		PTR	C_LABEL(sys_dup)
		PTR	C_LABEL(sys_pipe)
		PTR	C_LABEL(sys_times)
		PTR	C_LABEL(sys_prof)
		PTR	C_LABEL(sys_brk)		/* 45 */
		PTR	C_LABEL(sys_setgid)
		PTR	C_LABEL(sys_getgid)
		PTR	C_LABEL(sys_signal)
		PTR	C_LABEL(sys_geteuid)
		PTR	C_LABEL(sys_getegid)		/* 50 */
		PTR	C_LABEL(sys_acct)
		PTR	C_LABEL(sys_phys)
		PTR	C_LABEL(sys_lock)
		PTR	C_LABEL(sys_ioctl)
		PTR	C_LABEL(sys_fcntl)		/* 55 */
		PTR	C_LABEL(sys_mpx)
		PTR	C_LABEL(sys_setpgid)
		PTR	C_LABEL(sys_ulimit)
		PTR	C_LABEL(sys_olduname)
		PTR	C_LABEL(sys_umask)		/* 60 */
		PTR	C_LABEL(sys_chroot)
		PTR	C_LABEL(sys_ustat)
		PTR	C_LABEL(sys_dup2)
		PTR	C_LABEL(sys_getppid)
		PTR	C_LABEL(sys_getpgrp)		/* 65 */
		PTR	C_LABEL(sys_setsid)
		PTR	C_LABEL(sys_sigaction)
		PTR	C_LABEL(sys_sgetmask)
		PTR	C_LABEL(sys_ssetmask)
		PTR	C_LABEL(sys_setreuid)		/* 70 */
		PTR	C_LABEL(sys_setregid)
		PTR	C_LABEL(sys_sigsuspend)
		PTR	C_LABEL(sys_sigpending)
		PTR	C_LABEL(sys_sethostname)
		PTR	C_LABEL(sys_setrlimit)		/* 75 */
		PTR	C_LABEL(sys_getrlimit)
		PTR	C_LABEL(sys_getrusage)
		PTR	C_LABEL(sys_gettimeofday)
		PTR	C_LABEL(sys_settimeofday)
		PTR	C_LABEL(sys_getgroups)		/* 80 */
		PTR	C_LABEL(sys_setgroups)
		PTR	C_LABEL(sys_select)
		PTR	C_LABEL(sys_symlink)
		PTR	C_LABEL(sys_lstat)
		PTR	C_LABEL(sys_readlink)		/* 85 */
		PTR	C_LABEL(sys_uselib)
		PTR	C_LABEL(sys_swapon)
		PTR	C_LABEL(sys_reboot)
		PTR	C_LABEL(sys_readdir)
		PTR	C_LABEL(sys_mmap)		/* 90 */
		PTR	C_LABEL(sys_munmap)
		PTR	C_LABEL(sys_truncate)
		PTR	C_LABEL(sys_ftruncate)
		PTR	C_LABEL(sys_fchmod)
		PTR	C_LABEL(sys_fchown)		/* 95 */
		PTR	C_LABEL(sys_getpriority)
		PTR	C_LABEL(sys_setpriority)
		PTR	C_LABEL(sys_profil)
		PTR	C_LABEL(sys_statfs)
		PTR	C_LABEL(sys_fstatfs)		/* 100 */
		PTR	C_LABEL(sys_ioperm)
		PTR	C_LABEL(sys_socketcall)
		PTR	C_LABEL(sys_syslog)
		PTR	C_LABEL(sys_setitimer)
		PTR	C_LABEL(sys_getitimer)		/* 105 */
		PTR	C_LABEL(sys_newstat)
		PTR	C_LABEL(sys_newlstat)
		PTR	C_LABEL(sys_newfstat)
		PTR	C_LABEL(sys_uname)
		PTR	C_LABEL(sys_iopl)		/* 110 */
		PTR	C_LABEL(sys_vhangup)
		PTR	C_LABEL(sys_idle)
		PTR	C_LABEL(sys_vm86)
		PTR	C_LABEL(sys_wait4)
		PTR	C_LABEL(sys_swapoff)		/* 115 */
		PTR	C_LABEL(sys_sysinfo)
		PTR	C_LABEL(sys_ipc)
		PTR	C_LABEL(sys_fsync)
		PTR	C_LABEL(sys_sigreturn)
		PTR	C_LABEL(sys_clone)		/* 120 */
		PTR	C_LABEL(sys_setdomainname)
		PTR	C_LABEL(sys_newuname)
		PTR	0 #C_LABEL(sys_modify_ldt)
		PTR	C_LABEL(sys_adjtimex)
		PTR	C_LABEL(sys_mprotect)		/* 125 */
		PTR	C_LABEL(sys_sigprocmask)
		PTR	C_LABEL(sys_create_module)
		PTR	C_LABEL(sys_init_module)
		PTR	C_LABEL(sys_delete_module)
		PTR	C_LABEL(sys_get_kernel_syms)	/* 130 */
		PTR	C_LABEL(sys_quotactl)
		PTR	C_LABEL(sys_getpgid)
		PTR	C_LABEL(sys_fchdir)
		PTR	C_LABEL(sys_bdflush)
		PTR	C_LABEL(sys_sysfs)		/* 135 */
		PTR	C_LABEL(sys_personality)
		PTR	0			/* for afs_syscall */
		PTR	C_LABEL(sys_setfsuid)
		PTR	C_LABEL(sys_setfsgid)
		PTR	C_LABEL(sys_llseek)		/* 140 */
		.space	(NR_syscalls-140)*4

		.bss
		EXPORT(IRQ_vectors)
		.fill	16,4,0

